
#include <zephyr/kernel.h>
#include <zephyr/sys/printk.h>
#include <zephyr/sys/crc.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <esb.h>
#include "FPGApacket.h"

// Global structures
const char * Fpga_put_rslt_txt[] = {"1stPartOK", "LastPartOK", "PartOK", "AllPartsOK", "CrcErr\r\n", "FpgaHeaderIdErr\r\n", "EsbIdxErr\r\n"};   // short texts
FPacket_t Fpga_packets[FPGA_PACKET_NUM];
uint16_t esb_payload_count = 0;

// ****************************************************************************
// todo: prepracovat na pointrovu verziu
void fill_f_packets(void) {
    for (uint16_t i = 0; i < FPGA_PACKET_NUM; i++) {
        Fpga_packets[i].headerID = 0xAA00 + i;
        Fpga_packets[i].timeStampMSB = i;
        Fpga_packets[i].timeStampLSB = i + 1;
        Fpga_packets[i].measurementInfo = 0x1234;
        Fpga_packets[i].payloadSize = FPGA_PAYLOAD_SIZE;
        Fpga_packets[i].payloadID = i;

        // Fill payload with incremental values
        for (uint16_t j = 0; j < FPGA_PAYLOAD_SIZE; j++) {
            Fpga_packets[i].payload[j] = (uint16_t)(i + j);
        }

        Fpga_packets[i].footerID = 0xCCDD;
        Fpga_packets[i].CRC16 = fpga_packet_crc_calc(&Fpga_packets[i]);
    }
}

// ****************************************************************************
uint8_t fill_esb_packets(struct esb_payload * esb_payload_ptr) {
    uint16_t esb_index = 0;
    uint16_t fpga_headerID;                                 // each ESB will have an FPGA header ID
    uint8_t  esb_sub_index;                                 // ESB packet numbering within an FPGA packet
    uint16_t data_size;
    uint16_t data_shift;

    for (uint16_t i = 0; i < FPGA_PACKET_NUM; i++) {
        uint8_t *fpga_packetPtr = (uint8_t*)&Fpga_packets[i] + FPGA_HEADER_SIZE;    // Convert entire struct to uint8_t pointer
        fpga_headerID = Fpga_packets[i].headerID;           // catch the FPGA header ID
        int16_t remainingBytes = FPGA_PACKET_SIZE - FPGA_HEADER_SIZE;         // Total bytes to send (without FPGA header ID)
        uint8_t pipe = i % 6;                               // Cycle through pipes
//        uint8_t pid = 0;                                   // Packet ID for ESB packet sequence
        esb_sub_index = 0;
#if TEST_FPGA_PRINT_PACKETS
        print_data_array(fpga_packetPtr, remainingBytes, 100);
#endif

        while (remainingBytes > FPGA_INFO_SIZE && esb_index < ESB_PACKET_NUM) {
            uint16_t chunkSize = (remainingBytes > CONFIG_ESB_MAX_PAYLOAD_LENGTH) ? CONFIG_ESB_MAX_PAYLOAD_LENGTH : remainingBytes;

            // Fill ESB payload structure
            esb_payload_ptr->pipe = pipe;
            esb_payload_ptr->rssi = -42;                    // Placeholder RSSI value
            esb_payload_ptr->noack = 0;                     // Enable / Disable acknowledgments
//            esb_payload_ptr->pid = pid++;                   // Increment PID for each ESB fragment
            if(chunkSize > (CONFIG_ESB_MAX_PAYLOAD_LENGTH - FPGA_INFO_SIZE) ) {
                data_size = CONFIG_ESB_MAX_PAYLOAD_LENGTH - FPGA_INFO_SIZE;
            } else {
                data_size = chunkSize;
            }
            esb_payload_ptr->length = data_size + FPGA_INFO_SIZE;
#if TEST_FPGA_PRINT_DEBUG
            printk("i=%d, remainingBytes=%d, chunkSize=%d, fpga_packetPtr=%p, data_size=%d\r\n", i, remainingBytes, chunkSize, fpga_packetPtr, data_size);
#endif
            memcpy(&esb_payload_ptr->data[BATCH_FPGA_HEADER_IDX], (uint8_t *)&fpga_headerID, FPGA_HEADER_SIZE);     // first two data positions are FPGA header ID
            esb_payload_ptr->data[BATCH_ESB_SUBIND_IDX] = esb_sub_index;                            // 3th data position is ESB subindex within the FPGA packet
            memcpy(&esb_payload_ptr->data[BATCH_DATA_IDX], fpga_packetPtr, data_size );  // data
            data_shift = chunkSize - FPGA_INFO_SIZE;
 
            // Update pointers
            esb_sub_index++;
            fpga_packetPtr += data_shift;
            remainingBytes -= data_shift;
#if TEST_FPGA_PRINT_DEBUG
            print_packet_esb_single(esb_payload_ptr, esb_index, true);
#endif
            esb_index++;
            esb_payload_ptr++;
//            printk("i=%d, remainingBytes=%d, chunkSize=%d, fpga_packetPtr=%p\r\n\r\n", i, remainingBytes, chunkSize, fpga_packetPtr);
        }
    }
    esb_payload_count = esb_index;
    printk("***** Created %d total ESB payload(s). *****\n", esb_payload_count);

    return esb_payload_count;
}
// ****************************************************************************
FPGA_PUT_RESULT_e fpga_put_esb2fpga(FPacket_t * fpga_ptr, struct esb_payload * esb_ptr) {
    static uint16_t fpga_head_id_last = 0;                              // STATIC Current FPGA packet header ID
    static int8_t esb_idx_last = ESB_SUBIDX_NOINIT;                     // STATIC ESB index within the FPGA packet
    static uint8_t * fpga_dst_ptr = 0;                                  // STATIC uint8 pointer to FPGA destination

    FPGA_PUT_RESULT_e result_ret = PART_FIRST_OK;
    uint16_t fpga_head_id_received = *((uint16_t *)&esb_ptr->data[BATCH_FPGA_HEADER_IDX]);     // get FPGA header ID
    uint16_t esb_idx_received = esb_ptr->data[BATCH_ESB_SUBIND_IDX];      // get ESB subindex

    switch(esb_idx_last) {
        case ESB_SUBIDX_NOINIT:                                         // Q: Not yet initialized?
            if(esb_idx_received == ESB_SUBIDX_FIRST) {                  // Y: Q: Is the received ESB packet the first one?
                fpga_head_id_last = fpga_head_id_received;              // catch FPGA header ID
                memset((uint8_t *)fpga_ptr, 0, FPGA_PACKET_SIZE);              // clear the FPGA packet

                result_ret = PART_FIRST_OK;                             // indication
            } else {    
                result_ret = ESB_IDX_ERR;                               // N: error indication
            }
            break;

        default:
            if(fpga_head_id_last == fpga_head_id_received) {            // Y: Q: Is the ESB package from the same FPGA as the one being developed?
                if(esb_idx_received == esb_idx_last + 1) {              // Y: Is the received ESB packet the last one?
                    if(esb_idx_received == ESB_SUBIDX_LAST) {
                        result_ret = PART_LAST_OK;    
                    } else {
                        result_ret = PART_OK;
                    }
                } else {
                    result_ret = ESB_IDX_ERR;
                }
            } else {
                result_ret = FPGA_HEADER_ID_ERR;
            }
            break;
    }

    if(result_ret == PART_FIRST_OK || result_ret == PART_LAST_OK || result_ret == PART_OK) {
        if(result_ret == PART_FIRST_OK) {
            fpga_ptr->headerID = fpga_head_id_received;
            fpga_dst_ptr = (uint8_t *)fpga_ptr + FPGA_HEADER_SIZE;
        }

        uint8_t data_size = esb_ptr->length - FPGA_INFO_SIZE;
        memcpy(fpga_dst_ptr, &esb_ptr->data[BATCH_DATA_IDX], data_size);
        esb_idx_last++;                                     // 
        fpga_dst_ptr += data_size;
    }

#if TEST_FPGA_PUT_PRINT_STATES
    printk("fpga_head_id_last=0x%04x, esb_idx_last=%d, fpga_dst_ptr=%p\r\n", fpga_head_id_last, esb_idx_last, fpga_dst_ptr);
#endif

#if TEST_FPGA_PUT_PRINT_ESB    
    printk("esb_ptr->data[SET_DATA_IDX]:");
    print_data_array(&esb_ptr->data[SET_DATA_IDX], 20, 100);
#endif

    if(result_ret == PART_LAST_OK) {
#if TEST_FPGA_PUT_PRINT_FPGA
        print_data_array((uint8_t *) fpga_ptr, FPGA_PACKET_SIZE, 100);
#endif
        esb_idx_last = ESB_SUBIDX_NOINIT;
        fpga_head_id_last = 0;
        uint16_t fpga_crc_calc = fpga_packet_crc_calc(fpga_ptr);
        if(fpga_crc_calc == fpga_ptr->CRC16) {
            result_ret = PARTS_ALL_OK;
        } else {
            result_ret = CRC_ERR;
        }
    }

    return result_ret;
}

// ****************************************************************************
void print_packet_esb_single(struct esb_payload * esb_payload_ptr, uint8_t esb_packet_id, bool print_data_enable) {
    printf("Packet %02d | Pipe: %d | Length: %03d | RSSI: %d | PID: %d | Data: ",
        esb_packet_id, esb_payload_ptr->pipe, esb_payload_ptr->length, esb_payload_ptr->rssi, esb_payload_ptr->pid);

    if(print_data_enable) {
        for (uint8_t data_idx = 0; data_idx < esb_payload_ptr->length; data_idx++) {
            printf("%02X ", esb_payload_ptr->data[data_idx]);
        }
    }
    printf("\n");
}

// ****************************************************************************
void print_packet_esb_array(struct esb_payload * esb_payload_ptr, uint8_t esb_packet_num, bool print_data_enable) {
    for (uint16_t esb_idx = 0; esb_idx < esb_packet_num; esb_idx++, esb_payload_ptr++) {
        print_packet_esb_single(esb_payload_ptr, esb_idx, print_data_enable);
    }
}

// ****************************************************************************
void print_data_array(uint8_t * data_ptr, uint16_t data_size, uint8_t data_line_size) {
    for (uint16_t i = 0; i < data_size; i++, data_ptr++) {
        printk("%02x ", *data_ptr);
        if(i && i % data_line_size == 0) {
            printk("\r\n");
        }
    }
    printk("\r\n");
}

// ****************************************************************************
void compare_packets(struct esb_payload *esb_payload_ptr, struct esb_payload *esb_pattern_ptr, uint16_t index) {
    if (esb_payload_ptr->length != esb_pattern_ptr[index].length) {
        printf("Packets are not equal: Length mismatch h: %d, d: %d, I: %d\n",esb_payload_ptr->length, esb_pattern_ptr[index].length, index);
        return;
    }
    if (memcmp(esb_payload_ptr->data, esb_pattern_ptr[index].data, esb_payload_ptr->length) != 0) {
        printf("Packets are not equal: Data mismatch\n");
        return;
    }

    printf("Packets are equal! PKT index: %d, PKT len: %d\n", index, esb_pattern_ptr[index].length);
}

// ****************************************************************************
uint16_t fpga_packet_crc_calc(FPacket_t * fpga_packet_ptr) {
    return crc16_itu_t(FPGA_CRC_SEED, (uint8_t*)&(fpga_packet_ptr->headerID), FPGA_CRC_SRC_SIZE);
}