/*
 * Copyright (c) 2018 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/drivers/clock_control.h>
#include <zephyr/drivers/clock_control/nrf_clock_control.h>
#if defined(NRF54L15_XXAA)
	#include <hal/nrf_clock.h>
#endif /* defined(NRF54L15_XXAA) */
#include <zephyr/drivers/gpio.h>
#include <zephyr/irq.h>
#include <zephyr/logging/log.h>
#include <nrf.h>
#include <zephyr/kernel.h>
#include <zephyr/sys/crc.h>
#include <zephyr/types.h>
#include <dk_buttons_and_leds.h>
#if defined(CONFIG_CLOCK_CONTROL_NRF2)
	#include <hal/nrf_lrcconf.h>
#endif
#include "device_cfg.h"										// basic configuration
#include "driver_ESB.h"  									/* constants, macros and function for ESB service */
#include "FPGApacket.h"  									/* Declares fill_f_packets(), tx_payload[], etc. */

// ****************************************************************************
LOG_MODULE_REGISTER(main_prx, CONFIG_ESB_PRX_APP_LOG_LEVEL);
//                          Test management directives
#define TEST_PRINT_ESBRX_DATA	1       /* enable/disable - print the data from received ESB packet */
#define TEST_RECEIVING_FPGA_FAST	1	/* enable/disable - fast debug mode for FPGA receiving */

static ESBcontrol_t Esb_ctrl;								// ESB control structure

const uint8_t Base_addr_0[ESB_BASE_ADDR0_SIZE] = ESB_BASE_ADDR_0;
const uint8_t Base_addr_1[ESB_BASE_ADDR1_SIZE] = ESB_BASE_ADDR_1;
const uint8_t Addr_prefix[ESB_PREFIX_SIZE] = ESB_ADDR_PREFIX;
// ****************************************************************************
static void leds_update(uint8_t value)
{
	uint32_t leds_mask =
		(!(value % 8 > 0 && value % 8 <= 4) ? DK_LED1_MSK : 0) |
		(!(value % 8 > 1 && value % 8 <= 5) ? DK_LED2_MSK : 0) |
		(!(value % 8 > 2 && value % 8 <= 6) ? DK_LED3_MSK : 0) |
		(!(value % 8 > 3) ? DK_LED4_MSK : 0);

	dk_set_leds(leds_mask);
}
/* ------------------------------------------------------------------ */

#if defined(CONFIG_CLOCK_CONTROL_NRF)
int clocks_start(void)
{
	int err;
	int res;
	struct onoff_manager *clk_mgr;
	struct onoff_client clk_cli;

	clk_mgr = z_nrf_clock_control_get_onoff(CLOCK_CONTROL_NRF_SUBSYS_HF);
	if (!clk_mgr) {
		LOG_ERR("Unable to get the Clock manager");
		return -ENXIO;
	}

	sys_notify_init_spinwait(&clk_cli.notify);

	err = onoff_request(clk_mgr, &clk_cli);
	if (err < 0) {
		LOG_ERR("Clock request failed: %d", err);
		return err;
	}

	do {
		err = sys_notify_fetch_result(&clk_cli.notify, &res);
		if (!err && res) {
			LOG_ERR("Clock could not be started: %d", res);
			return res;
		}
	} while (err);

#if defined(NRF54L15_XXAA)
	/* MLTPAN-20 */
	nrf_clock_task_trigger(NRF_CLOCK, NRF_CLOCK_TASK_PLLSTART);
#endif /* defined(NRF54L15_XXAA) */

	LOG_DBG("HF clock started");
	return 0;
}

#elif defined(CONFIG_CLOCK_CONTROL_NRF2)

int clocks_start(void)
{
	int err;
	int res;
	const struct device *radio_clk_dev =
		DEVICE_DT_GET_OR_NULL(DT_CLOCKS_CTLR(DT_NODELABEL(radio)));
	struct onoff_client radio_cli;

	/** Keep radio domain powered all the time to reduce latency. */
	nrf_lrcconf_poweron_force_set(NRF_LRCCONF010, NRF_LRCCONF_POWER_DOMAIN_1, true);

	sys_notify_init_spinwait(&radio_cli.notify);

	err = nrf_clock_control_request(radio_clk_dev, NULL, &radio_cli);

	do {
		err = sys_notify_fetch_result(&radio_cli.notify, &res);
		if (!err && res) {
			LOG_ERR("Clock could not be started: %d", res);
			return res;
		}
	} while (err == -EAGAIN);

#if defined(NRF54L15_XXAA)
	/* MLTPAN-20 */
	nrf_clock_task_trigger(NRF_CLOCK, NRF_CLOCK_TASK_PLLSTART);
#endif /* defined(NRF54L15_XXAA) */

	LOG_DBG("HF clock started");

	return 0;
}

#else
BUILD_ASSERT(false, "No Clock Control driver");
#endif /* defined(CONFIG_CLOCK_CONTROL_NRF2) */

int main(void)
{
	int err;
	uint8_t esb_packet_no = 0;
	uint16_t fpga_in_crc = 0;								// CRC of received FPGA packet
//	int esb_read_status;						 				// return value from reading f-ction from esb_event handler - first initialization
	uint32_t dbg_cnt = 0;
	bool led_dbg_tgl = false;
	// int sem_ret;											// return value from semaphore
	// int fifo_ret;											// return value from FIFO
	// static struct esb_payload * tx_payload_ptr;				// pointer to ESB tx buffer
	static struct esb_payload esb_payload_tmp;				// temporary ESB payload - STATIC!!! (because pointer to this variable is argument in the function)
	static struct data_esb_item_t * data_esb_rx_item_ptr;	//
	static FPacket_t fpga_packet_dst;						// FPGA packet for received data - STATIC!!! (because pointer to this variable is argument in the function)
//	uint16_t fpga_packet_current_idx;						// index/headerID of current FPGA packet
	FPGA_PUT_RESULT_e fpga_put_rslt = PART_FIRST_OK;
	char uart_tx_buf[25];
	sprintf(uart_tx_buf, "%d.%02d.%03d", DVC_FW_VER_MAJOR, DVC_FW_VER_MINOR, DVC_FW_VER_PATCH);
	LOG_INF("%s, SW version: %s", DVC_NAME_TXT, uart_tx_buf);

    /* Start HF clock if needed. */
	err = clocks_start();
	if (err) {
        printk("clocks_start() failed, err=%d\n", err);
		return 0;
	}

	err = dk_leds_init();
	if (err) {
		LOG_ERR("LEDs initialization failed, err %d", err);
		return 0;
	}

    /* Initialize ESB driver. */
	ESBconfig_t drv_esb_cfg = {
		.base_addr_0_ptr = Base_addr_0,
		.base_addr_1_ptr = Base_addr_1,
		.addr_prefix_ptr = Addr_prefix,
		.mode = TRANSMIT_ESB_MODE,
		.ack_enable = TRANSMIT_ACK_CTRL,
	};

	err = drv_esb_init(&Esb_ctrl, &drv_esb_cfg);
	if (err) {
		LOG_ERR("ESB initialization failed, err %d", err);
		return 0;
	}

	LOG_INF("Initialization complete");

#if 0
	err = esb_write_payload(&tx_payload);
	if (err) {
		LOG_ERR("Write payload, err %d", err);
		return 0;
	}
#endif
	LOG_INF("Setting up for packet receiption");

	err = esb_start_rx();
	if (err) {
		LOG_ERR("RX setup failed, err %d", err);
		return 0;
	}
	LOG_DBG("Esb_ctrl.tx_sem_ptr=%p, Esb_ctrl.rx_fifo_ptr=%p, Esb_ctrl.tx_ready=%d", 
		Esb_ctrl.tx_sem_ptr, Esb_ctrl.rx_fifo_ptr, Esb_ctrl.tx_ready);


	while(1) {
//		data_esb_rx_item_ptr = k_fifo_get(Esb_ctrl.rx_fifo_ptr, K_MSEC(TRANSMIT_DELAY_FPGA_ARR_ms + TRANSMIT_WAIT_OVERLAP_ms));	// waiting to receive ESB packet from handler service
		k_sem_take(Esb_ctrl.rx_sem_ptr, K_MSEC(50));	// waits for RX done - via a semaphore controlled from the ESB handler

//		printk("State 1.\r\n");
//		 esb_read_status = esb_read_rx_payload(&esb_payload_tmp);		// reads the received ESB packet
//		if(fifo_ret != -EAGAIN) {

//		if(data_esb_rx_item_ptr != NULL) {
		if(1)	{
			int esb_read_status = esb_read_rx_payload(&esb_payload_tmp);		// reads the received ESB packet to temporary ESB packet
//			if(data_esb_rx_item_ptr->read_status == 0) {
			if(esb_read_status == 0) {
//				printk("State Data Received.\r\n");
//				memcpy(&esb_payload_tmp, &data_esb_rx_item_ptr->payload, data_esb_rx_item_ptr->payload.length + 5);		// skontolovat opodstatnenost cisla "5"

// 14.3.2025				esb_flush_rx();
//				print_packet_esb_single(&esb_payload_tmp, esb_payload_tmp.data[BATCH_ESB_SUBIND_IDX], TEST_PRINT_ESBRX_DATA);
				fpga_put_rslt = fpga_put_esb2fpga(&fpga_packet_dst, &esb_payload_tmp);
				dk_set_led_off(2);
#if TEST_RECEIVING_FPGA_FAST == 0
				if(fpga_put_rslt == PART_FIRST_OK) {
					printk("\r\nFpgaStatus: ");		// "printk" due to time performance
				}
				printk("%s,", Fpga_put_rslt_txt[(uint8_t)fpga_put_rslt]);		// "printk" due to time performance
#else			
				if(fpga_put_rslt == PART_FIRST_OK) {	// Q: Is received ESB packet first from FPGA packet?
//					printk("FPGA receiving started. \r\n");	// print separator line
				}
	#if TEST_FPGA_PUT_CHECK_CRC
				if(fpga_put_rslt == PARTS_ALL_OK || fpga_put_rslt == CRC_ERR || fpga_put_rslt == FPGA_HEADER_ID_ERR /*|| fpga_put_rslt == ESB_IDX_ERR*/) 
	#else					
				if(fpga_put_rslt == PART_LAST_OK || fpga_put_rslt == FPGA_HEADER_ID_ERR || fpga_put_rslt == ESB_IDX_ERR) 
	#endif
				{
				printk("%s\r\n", Fpga_put_rslt_txt[(uint8_t)fpga_put_rslt]);		// "printk" due to time performance
				}
#endif
				// leds_update(esb_payload_tmp.data[0]);
				// } else {
				// LOG_ERR("Error while reading rx packet");
			}	/* if(data_esb_rx_item_ptr->read_status) */
		} else {	// if(data_esb_rx_item_ptr...)
			dk_set_led_off(2);
			LOG_DBG("\r\nPRX timeout");
		}
//		printk("State Counts.\r\n");
		dbg_cnt++;
		//if(dbg_cnt % 100000 == 0) 
		{
			led_dbg_tgl = !led_dbg_tgl;
		}

		if(led_dbg_tgl) {
			dk_set_led_on(1);
		} else {
			dk_set_led_off(1);
		}
	}	/* end of while(1) */

	/* return to idle thread */
	printk("Ranaway from main.");

	return 0;
}	/* enf of main() */
