/*
 * Copyright (c) 2018 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/drivers/clock_control.h>
#include <zephyr/drivers/clock_control/nrf_clock_control.h>
#if defined(NRF54L15_XXAA)
	#include <hal/nrf_clock.h>
#endif /* defined(NRF54L15_XXAA) */
#include <zephyr/drivers/gpio.h>
#include <zephyr/irq.h>
#include <zephyr/logging/log.h>
#include <nrf.h>
#include <zephyr/kernel.h>
#include <zephyr/types.h>
#include <dk_buttons_and_leds.h>
#if defined(CONFIG_CLOCK_CONTROL_NRF2)
	#include <hal/nrf_lrcconf.h>
#endif
#include "device_cfg.h"										// basic configuration
#include "driver_ESB.h"  									/* constants, macros and function for ESB service */
#include "FPGApacket.h"  									/* Declares fill_f_packets(), tx_payload[], etc. */

// ****************************************************************************
LOG_MODULE_REGISTER(main_ptx, CONFIG_ESB_PTX_APP_LOG_LEVEL);

static ESBcontrol_t Esb_ctrl;								// ESB control structure
struct esb_payload tx_payload[ESB_PACKET_NUM];
const uint8_t Base_addr_0[ESB_BASE_ADDR0_SIZE] = ESB_BASE_ADDR_0;
const uint8_t Base_addr_1[ESB_BASE_ADDR1_SIZE] = ESB_BASE_ADDR_1;
const uint8_t Addr_prefix[ESB_PREFIX_SIZE] = ESB_ADDR_PREFIX;

/* We increment this in the event handler (optional). */
static uint32_t sequence_number = 0;

/* For measuring total bytes sent and any error attempts. */
static uint32_t esb_data_sent         = 0;

/* For measuring total bytes sent and any error attempts. */
static uint32_t send_err_attempts = 0;

// ****************************************************************************
static void leds_update(uint8_t value)
{
	uint32_t leds_mask =
		(!(value % 8 > 0 && value % 8 <= 4) ? DK_LED1_MSK : 0) |
		(!(value % 8 > 1 && value % 8 <= 5) ? DK_LED2_MSK : 0) |
		(!(value % 8 > 2 && value % 8 <= 6) ? DK_LED3_MSK : 0) |
		(!(value % 8 > 3) ? DK_LED4_MSK : 0);

	dk_set_leds(leds_mask);
}
/* ------------------------------------------------------------------ */

#if defined(CONFIG_CLOCK_CONTROL_NRF)
int clocks_start(void)
{
	int err;
	int res;
	struct onoff_manager *clk_mgr;
	struct onoff_client clk_cli;

	clk_mgr = z_nrf_clock_control_get_onoff(CLOCK_CONTROL_NRF_SUBSYS_HF);
	if (!clk_mgr) {
		LOG_ERR("Unable to get the Clock manager");
		return -ENXIO;
	}

	sys_notify_init_spinwait(&clk_cli.notify);

	err = onoff_request(clk_mgr, &clk_cli);
	if (err < 0) {
		LOG_ERR("Clock request failed: %d", err);
		return err;
	}

	do {
		err = sys_notify_fetch_result(&clk_cli.notify, &res);
		if (!err && res) {
			LOG_ERR("Clock could not be started: %d", res);
			return res;
		}
	} while (err);

#if defined(NRF54L15_XXAA)
	/* MLTPAN-20 */
	nrf_clock_task_trigger(NRF_CLOCK, NRF_CLOCK_TASK_PLLSTART);
#endif /* defined(NRF54L15_XXAA) */

	LOG_DBG("HF clock started");
	return 0;
}

#elif defined(CONFIG_CLOCK_CONTROL_NRF2)

int clocks_start(void)
{
	int err;
	int res;
	const struct device *radio_clk_dev =
		DEVICE_DT_GET_OR_NULL(DT_CLOCKS_CTLR(DT_NODELABEL(radio)));
	struct onoff_client radio_cli;

	/** Keep radio domain powered all the time to reduce latency. */
	nrf_lrcconf_poweron_force_set(NRF_LRCCONF010, NRF_LRCCONF_POWER_DOMAIN_1, true);

	sys_notify_init_spinwait(&radio_cli.notify);

	err = nrf_clock_control_request(radio_clk_dev, NULL, &radio_cli);

	do {
		err = sys_notify_fetch_result(&radio_cli.notify, &res);
		if (!err && res) {
			LOG_ERR("Clock could not be started: %d", res);
			return res;
		}
	} while (err == -EAGAIN);

#if defined(NRF54L15_XXAA)
	/* MLTPAN-20 */
	nrf_clock_task_trigger(NRF_CLOCK, NRF_CLOCK_TASK_PLLSTART);
#endif /* defined(NRF54L15_XXAA) */

	LOG_DBG("HF clock started");

	return 0;
}

#else
BUILD_ASSERT(false, "No Clock Control driver");
#endif /* defined(CONFIG_CLOCK_CONTROL_NRF2) */

int main(void)
{
	int err;
	uint8_t esb_idx;
	uint8_t esb_packets_cnt;
    uint32_t time_start_ms, time_end_ms, time_delta_ms;
	static struct esb_payload * tx_payload_ptr;				// pointer to ESB tx buffer
	char uart_tx_buf[25];
	sprintf(uart_tx_buf, "%d.%02d.%03d", DVC_FW_VER_MAJOR, DVC_FW_VER_MINOR, DVC_FW_VER_PATCH);
	LOG_INF("%s, SW version: %s", DVC_NAME_TXT, uart_tx_buf);

    /* Start HF clock if needed. */
	err = clocks_start();
	if (err) {
        printk("clocks_start() failed, err=%d\n", err);
		return 0;
	}

	err = dk_leds_init();
	if (err) {
		LOG_ERR("LEDs initialization failed, err %d", err);
		return 0;
	}

    /* Initialize ESB driver. */
	ESBconfig_t drv_esb_cfg = {
		.base_addr_0_ptr = Base_addr_0,
		.base_addr_1_ptr = Base_addr_1,
		.addr_prefix_ptr = Addr_prefix,
		.mode = TRANSMIT_ESB_MODE,
		.ack_enable = TRANSMIT_ACK_CTRL,
	};

	err = drv_esb_init(&Esb_ctrl, &drv_esb_cfg);
	if (err) {
		LOG_ERR("ESB initialization failed, err %d", err);
		return 0;
	}

	LOG_INF("Initialization complete");
	LOG_INF("Sending test packet");
    /* Build large FPGA packets & then split them into ESB fragments. */
    fill_f_packets();
    esb_packets_cnt = fill_esb_packets(tx_payload);
    print_packet_esb_array(&tx_payload, ESB_PACKET_NUM, TEST_PRINT_ESBTX_DATA);  						/* For debugging only; comment out if too verbose. */
	tx_payload_ptr = &tx_payload[0];
	// tx_payload_ptr->noack = false;


	
	while (1) {
		LOG_DBG("***** FPGA packet sending has begun. *****");

		for(esb_idx = 0, tx_payload_ptr = &tx_payload[0], esb_data_sent = 0; esb_idx < esb_packets_cnt; esb_idx++, tx_payload_ptr++) {
			// if (Esb_ctrl.tx_ready) {							// old blocking: waits for TX ready via a global variable controlled from the ESB handler
#if TRANSMIT_ACK_CTRL
				k_sem_take(Esb_ctrl.tx_sem_ptr, K_MSEC(50));	// waits for TX ready - via a semaphore controlled from the ESB handler
#else
				if(esb_idx > 0) {								// skip the very first transmission
//					while(esb_tx_full()) {						// wait for TX ready
					while(esb_is_idle()) {						// wait for TX ready
					}
				}
#endif
				dk_set_led_off(1);
				Esb_ctrl.tx_ready = false;
				tx_payload_ptr->noack = !Esb_ctrl.ack_enable;
				esb_flush_tx();
				esb_data_sent += tx_payload_ptr->length + 5;	// 5 - zadefinovat ako konstantu!!! - este to vsak zvazit, ci vobec
//				LOG_DBG("Sending ESB with idx=%d", esb_idx);
				if(esb_idx == 0) {								// catch timer before first one transmit
					time_start_ms = k_uptime_get_32();
					dk_set_led_on(3);
				}
				dk_set_led_on(0);
				err = esb_write_payload(tx_payload_ptr);
				if (err) {
					LOG_ERR("Payload write failed, err %d", err);
				}

#if TRANSMIT_DELAY_ESB_ms
 			k_sleep(K_MSEC(TRANSMIT_DELAY_ESB_ms));
#endif
#if TRANSMIT_DELAY_FPGA_EXTRA_ms
			if( esb_idx && (esb_idx + 1) % ESB_PACKET_PER_FPGA_PACKET == 0) {		// additional delay after FPGA termination
				k_sleep(K_MSEC(TRANSMIT_DELAY_FPGA_EXTRA_ms));
			}
#endif		
		}
		dk_set_led_off(3);
        time_end_ms = k_uptime_get_32();
		LOG_DBG("\n----- Sending completted. -----\n");
        time_delta_ms  = time_end_ms - time_start_ms;
		uint32_t esb_data_rate_kbps = esb_data_sent * 8 / MAX(1, time_delta_ms);	// [Byte] * 8 bits/Byte = [bits], [bits] / [ms] = [kbps], MAX() ~ avoid division by zero

        LOG_DBG("Sent total %u bytes in %u ms. Approx throughput: %u kbps.\n", esb_data_sent, time_delta_ms,
            esb_data_rate_kbps);
        LOG_DBG("Failed send attempts: %u", send_err_attempts);
        LOG_DBG("Final sequence_number: %u", sequence_number);

		k_sleep(K_MSEC(TRANSMIT_DELAY_FPGA_ARR_ms));
	}
}

#if 0
	while (1) {
		if (ready) {
			ready = false;
			esb_flush_tx();
			leds_update(tx_payload.data[1]);

			err = esb_write_payload(&tx_payload);
			if (err) {
				LOG_ERR("Payload write failed, err %d", err);
			}
			tx_payload.data[1]++;
		}

		k_sleep(K_MSEC(1000));
	}
#endif